

// 22/02/2021 P.R.

#include <SPI.h>
#include <Ethernet.h>
#include <RFID.h>
//#include <LiquidCrystal.h>  // SPI INTERFACE
#include <LiquidCrystal_I2C.h>
#include <EEPROM.h>
#include <SD.h>
#include <EthernetUdp.h>
#include <Time.h>
#include <Wire.h>  
#include <DS1307RTC.h>
#include <ArduinoJson.h>

#define SD_CS 4
#define RFID_IO1 40
#define RFID_RST1 41
#define RFID_IO2 38
#define RFID_RST2 39

#define buzzer 5              // speaker or buzzer on pin 
#define buzzer_gnd 2
#define led_green 48          // green LED 
#define led_green_gnd 49    
#define led_red 46            // led di errore
#define led_red_gnd 47
#define clock_gnd 22

String versioneSw = "Versione sw.1.00";
String dataSw = " del 19-02-2021";

char dayOfWeek[9][4] = {"", "DOM", "LUN", "MAR", "MER", "GIO", "VEN", "SAB"};
time_t t;
int displayAtSecond;

// timeZoneOffset = (Time Zone) * 3600L eg. (+8) * 3600L = 28800L for Taipei, Taiwan
const long timeZoneOffset = 3600L; 
// sync to NTP server every "ntpSyncInterval" seconds, set to 1 hour or more to be reasonable
unsigned long ntpSyncInterval;
// adjust the sync latency with computer NTP client in seconds 
unsigned long syncLatency = 0;
// keep track of how long ago we updated the NTP server
unsigned long ntpLastUpdate = 0;

IPAddress timeServer; 
const int NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[NTP_PACKET_SIZE]; // buffer to hold incoming and outgoing packets
// A UDP instance to let us send and receive packets over UDP
EthernetUDP Udp;

LiquidCrystal_I2C lcd(0x27,16,2);
//LiquidCrystal lcd(8,9,3,11,12,13);  //SPI INTERFACE
RFID RC522_IN(RFID_IO1, RFID_RST1); 
RFID RC522_OUT(RFID_IO2, RFID_RST2);

byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xE2 };
//IPAddress dnServer(8,8,8,8);
IPAddress dnServer;
IPAddress gateway;
IPAddress subnet;
IPAddress ip;
unsigned int localPort = 8888;              // Local UDP clock packets port
EthernetClient client;
char serverPingName[38];                    //(40 - (1°_char_dimensione) -1)
int serverPingPort = 80;
char serverCommandName[38]; 
int serverCommandPort;
String urlPost; 
const String fileTimbr = "timbr.txt";
File myFile;

const String configfilename = "config.jsn"; // <- SD library uses 8.3 filenames
File fileconfig; 

int risultatoDB;                            // check scrittura su mysql
int risultatoSD;                            // check scrittura su MicroSD

const int EEPROM_DB = 0;                    // indirizzo 0: Abilito scrittura su DB
const int EEPROM_SD = 1;                    // indirizzo 1: Abilito scrittura su SD card
const int EEPROM_CHK_UT = 2;                // check dell'esistenza dell'utente su db 
const int EEPROM_DEC_HEX = 3;               // Switch dei tag tra dec e hex
const int EEPROM_SEDE = 4;                  // codice Sede 
const int EEPROM_SERVERPINGNAME_ST = 10;    // addr 10-49    - il primo byte contiene la lunghezza della stringa max(39-1 byte)
const int EEPROM_NTP_SYNC_INTERVAL_ST = 50; // addr 50-59    - il primo byte contiene la lunghezza della stringa Max(9-1 byte)
const int EEPROM_SERVERCOMMANDNAME_ST = 60; // addr 60-99    - il primo byte contiene la lunghezza della stringa max(39-1 byte)
const int EEPROM_URLPOST_ST = 100;          // addr 100-139  - il primo byte contiene la lunghezza della stringa max(39-1 byte)
const int EEPROM_SERVERCOMMANDPORT_ST = 150;// addr 150-154  - il primo byte contiene la lunghezza della stringa max(4-1 byte)
const int EEPROM_IP_ST = 160;               // addr 160-179  - il primo byte contiene la lunghezza della stringa max(19-1 byte)
const int EEPROM_SUB_ST = 180;              // addr 180-199  - il primo byte contiene la lunghezza della stringa max(19-1 byte)
const int EEPROM_GAT_ST = 200;              // addr 200-219  - il primo byte contiene la lunghezza della stringa max(19-1 byte)
const int EEPROM_DNS_ST = 220;              // addr 220-239  - il primo byte contiene la lunghezza della stringa max(19-1 byte)
const int EEPROM_NTP_ST = 240;              // addr 240-259  - il primo byte contiene la lunghezza della stringa max(19-1 byte)

byte ee_db;                   // variabili per gli ind eeprom
byte ee_sd;   
byte ee_chk_ut;               // mysql enable/disable user check on mysql 
byte ee_hex_dec;              // read tag in dec format or hex format
byte ee_sede;                 
String sede3ch;               // uso sede con 3 char x formato standard programmi paghe
String tag;                   // id tag 13,65Mhz
String op;                    // ingresso:1 uscita:0
String dataop;                // date and time to log

//-------------------------------------------------------------------------
// SETUP SETUP SETUP SETUP SETUP SETUP SETUP SETUP SETUP SETUP SETUP SETUP 
//-------------------------------------------------------------------------
   
void setup() {
  pinMode(clock_gnd, OUTPUT);
  digitalWrite(clock_gnd, LOW);  
  pinMode(led_green, OUTPUT);
  pinMode(led_green_gnd, OUTPUT);
  pinMode(led_red, OUTPUT);
  pinMode(led_red_gnd, OUTPUT);
  pinMode(buzzer_gnd, OUTPUT);
  digitalWrite(led_green_gnd, LOW);
  digitalWrite(led_red_gnd, LOW);
  digitalWrite(buzzer_gnd, LOW);
   
  Serial.begin(9600);
  delay(200);
  Serial.print(F("Start... "));  
  //lcd.begin(16, 2); //SPI INTERFACE
  lcd.begin();  
  lcd.setCursor(0,0);
 
    // la SD va inizializzata prima della ethernet, altrimenti
    // con il reset non pinga più il server e va tolta l'alimentazione e rimessa. 
    if (!SD.begin(SD_CS)) {
      Serial.println(F("Err. SD!")); 
      lcd.clear();
      lcd.print(F("Err. SD!"));   
    }
    else {
      Serial.println(F("SD ok!"));
      lcd.clear();
      lcd.print(F("SD ok!")); 
    }  
    delay(500);
  //inizializza_defaults();  // per i vari test
  // Carico la configurazione dal file config.json 
  loadConfiguration();
  sede3ch="00";
  sede3ch.concat(ee_sede);  

  Ethernet.begin(mac, ip, dnServer, gateway, subnet);  
  delay(1000);
  Serial.print(Ethernet.localIP());
  Serial.println(F(" eth ready..."));
  ip = Ethernet.localIP();
  delay(500); 
  Udp.begin(localPort);
  delay(1000);
  lcd.clear();
  // Wire.begin();
  setSyncProvider(RTC.get);   // the function to get the time from the RTC
  if(timeStatus()!= timeSet) 
     lcd.print(F("Unable to sync"));
  else
    lcd.print(F("Sync system time "));
  
  displayAtSecond = second();
  delay(2000);
  lcd.clear();
  
  SPI.begin();  // run SPI library first; if not, RFID will not work
  RC522_IN.init();
  RC522_OUT.init();
  tone(buzzer, 1200, 100);
  delay(170);
  tone(buzzer, 1200, 100);
  delay(170);
  tone(buzzer, 1200, 100);
  delay(170);
  tone(buzzer, 1200, 100); 
  delay(750);

  lcd.clear();
  lcd.print(Ethernet.localIP());
  lcd.setCursor(0,1);
  if (Ethernet.hardwareStatus() == EthernetNoHardware) {
    Serial.println(F("Errore ETH."));
    lcd.print(F("    Errore ETH"));
  }
  if (Ethernet.hardwareStatus() == EthernetW5100) {
    Serial.println(F("W5100"));
    lcd.print(F("W5100"));
  }
  lcd.clear(); 
  lcd.print(versioneSw);
  lcd.setCursor(0,1);
  lcd.print(dataSw);
  Serial.println(versioneSw+dataSw);
  delay(1000);
  Serial.println(F("Ping svr.."));
  lcd.clear(); 
  lcd.print(F("Ping svr..."));
  delay(1000);
    
  if (client.connect(serverPingName, serverPingPort)) {
    Serial.println(F("Ping svr ok"));
    lcd.clear(); 
    lcd.print(F("Ping svr ok"));
    client.stop();
  }
  else {
    Serial.println(F("Ping server error"));
    lcd.clear(); 
    lcd.print(F("Ping svr error"));
  }
  
  delay(1000);
  lcd.clear(); 
}

//-------------------------------------------------------------------------
// MAIN LOOP MAIN LOOP MAIN LOOP MAIN LOOP MAIN LOOP MAIN LOOP MAIN LOOP 
//-------------------------------------------------------------------------
void loop() {
  
  tag = ""; 
  op = "";
  dataop = "";   

  // sync orologio con NTP
  if ((now() - ntpLastUpdate) >= ntpSyncInterval) {
    ntpSyncDS1307();
    // clear seconds displayed once a second when sync is needed, cause seconds blink to notify checking the network status
    lcd.setCursor(9, 1);
    lcd.print(F("       "));
  } 
  // for LCD shield to disp date, day of the week, time and temperature once a second
  t = now();
  if (displayAtSecond != second(t)) { 
    digitalClockDisplay(); 
    displayAtSecond = second(t); 
  }
   
  // Lettura - Timbro l'ingresso
  if (RC522_IN.isCard())  // check per la lettura del tag. 
  {       
    RC522_IN.readCardSerial();
    Serial.println(F("Tag IN: "));
    if(ee_hex_dec==0) {
      for (byte i=0; i<4; i++) {      
        if (RC522_IN.serNum[i] < 0x10) tag += "0";
        tag += String(RC522_IN.serNum[i], HEX);
      } 
    }
    else {
      for (byte i=0; i<4; i++) {       
        if (RC522_IN.serNum[i] < 0x10) tag += "0";
        tag += String(RC522_IN.serNum[i], DEC);
      } 
    }    
    tone(buzzer, 1200, 100); 
    Serial.println(tag);
    op="1";               //entrata
    dataop = setDataop();        
    if(ee_db==1)      
      insertToDb(sede3ch,tag,op,dataop);    // Tento di scrivere sul db il tag letto        
    if(ee_sd==1)
      insertToSD(sede3ch,tag,op,dataop); 
    delay(500);            
  }
  
  // Lettura - Timbro l'uscita
  if (RC522_OUT.isCard())  // check per la lettura del tag. 
  {     
    RC522_OUT.readCardSerial();
    Serial.println(F("Tag Out: "));
    if(ee_hex_dec==0) {
      for (byte i=0; i<4; i++) {        
        if (RC522_OUT.serNum[i] < 0x10) tag += "0";
        tag += String(RC522_OUT.serNum[i], HEX);
      } 
    }
    else {
      for (byte i=0; i<4; i++) {        
        if (RC522_OUT.serNum[i] < 0x10) tag += "0";
        tag += String(RC522_OUT.serNum[i], DEC);
      } 
    }   
    tone(buzzer, 1200, 100); 
    lcd.clear(); 
    lcd.print(tag);
    Serial.println(tag);
    delay(500);
    op = "0";                 // uscita
    dataop = setDataop();
    if(ee_db==1)      
      insertToDb(sede3ch,tag,op,dataop);    // Tento di scrivere sul db il tag letto        
    if(ee_sd==1)
      insertToSD(sede3ch,tag,op,dataop);  
    delay(500);
  }
}   // end loop

//---------------------------------------------------------------------------
// INSERIMENTO DB INSERIMENTO DB INSERIMENTO DB INSERIMENTO DB INSERIMENTO DB 
//---------------------------------------------------------------------------
void insertToDb(String f_sede3ch, String f_tag, String f_op, String f_dataop) {

  // compongo la stringa POST da passare alla pagina timbrature.php
  String postData = "&f_sede3ch="+f_sede3ch+"&f_tag=" + f_tag + "&f_op="+f_op+"&f_dataop="+f_dataop+"&ip="+ip+"&pwd=Rinald01"+"&chk_utente="+ee_chk_ut;
  // vado a controllare la risposta del server. 
  String readString; 
  Serial.print(F("connecting... "));
  if (client.connect(serverCommandName, serverCommandPort)) {
  Serial.println(F("connected"));
  client.print("POST ");
  client.print(urlPost);
  client.println(" HTTP/1.1");
  client.print("Host: ");
  client.println(serverCommandName);
  client.println("User-Agent: Arduino/1.0");
  client.println("Connection: close");
  client.println("Content-Type: application/x-www-form-urlencoded;");
  client.print("Content-Length: ");
  client.println(postData.length());
  client.println();
  client.println(postData);
   
  if (client) {   
    readString="";  
    while (client.connected()) {
    if (client.available()) {
        char c = client.read();
           readString += c;
    } } }     
    Serial.println(readString);
  } 
  else {
    readString="err";
    Serial.println(F("Errore di connessione"));
    lcd.clear(); 
    lcd.print(F("Errore di connessione"));
    Serial.println(readString);
  }

  // se tutto ok la risposta contiene il valore 34 Timbratura registrata
  // altrimenti contiene 36 Errore timbratura
  // in base a questo mostrerò sul display l'avvenuta registrazione o l'event errore.

   if(readString.indexOf("err") >=0) {    //connessione mancante  
      risultatoDB = 2;        
      lcd.clear();
      lcd.print(F("Errore di conness"));  
      delay(800);  
      lcd.clear();
      redLED(); 
   }
   
   if(readString.indexOf("34") >=0) {     // scrittura su db OK.
      risultatoDB = 1;                    //ok   
      lcd.clear();
      lcd.print(F("Mysql log OK!"));      
      delay(800);
      lcd.clear();
      greenLED();      
   }     
   
   if(readString.indexOf("36") >=0) {     //errore scrittura su DB
      risultatoDB = 0;                    //errore - utente non riconosciuto su db
      lcd.clear();
      lcd.print(F("Utente non trovato"));  
      delay(800); 
      lcd.clear(); 
      redLED();    
   }    
   Serial.print(F("Errori: 0:ut non trovato; 1:ok; 2:err.conn - "));
   Serial.println(risultatoDB);
   client.stop();
}

//---------------------------------------------------------------------------
// INSERIMENTO SD CARD INSERIMENTO SD CARD INSERIMENTO CD CARD INSERIMENTO SD 
// se la funzione è abilitata (ee_sd==1) registro sempre la timbratura, 
// senza alcun controllo sull'eventuale esistenza del tag.
//---------------------------------------------------------------------------
void insertToSD(String f_sede3ch, String f_tag, String f_op, String f_dataop) {
  // 0050030837fcd100001802211105
  String recordSd = f_sede3ch+"00"+f_tag+f_op+"0000"+f_dataop.substring(0,10);
  myFile = SD.open(fileTimbr, FILE_WRITE);
  if (myFile) {
    Serial.print(F("append to sd... "));        
    myFile.println(recordSd);       
    Serial.print(recordSd);    
    myFile.close();
    Serial.println(F(" SD write ok!"));
    lcd.clear();
    lcd.print(F("SD log OK!"));      
    delay(600);
    lcd.clear();
    digitalWrite(led_green, HIGH);
    tone(buzzer, 900, 80); 
    delay(100);
    digitalWrite(led_green, LOW);
    tone(buzzer, 1200, 80);     
  }
  else { 
    Serial.println(F("errore scrittura SD")); 
    lcd.clear();
    lcd.print(F("SD non pronta!"));      
    delay(600);
    lcd.clear();
    digitalWrite(led_red, HIGH);
    tone(buzzer, 1200, 80); 
    delay(100);
    digitalWrite(led_red, LOW);
    tone(buzzer, 900, 80); 
    
  }   
}

//---------------------------------------------------------------------------
// le funzioni che lampeggiano i led e buzzer 
//---------------------------------------------------------------------------
void redLED(){ // red LED on, green LED off
  for(int i=0;i<=1;i++) {
    int x = 1200-(i*300);
    digitalWrite(led_red, HIGH);
    tone(buzzer, x, 80); 
    delay(100);
  }
  digitalWrite(led_red, LOW);
  tone(buzzer, 300, 250); 
  delay(150);  
  for(int i=0;i<=2;i++) {
    digitalWrite(led_red, HIGH);
    delay(150);
    digitalWrite(led_red, LOW);
    delay(150);
  }
}

void greenLED(){ // red LED off, green LED on
  for(int i=0;i<=1;i++) {
    int x = 600+(i*300);
    digitalWrite(led_green, HIGH);
    tone(buzzer, x, 80); 
    delay(100);
  }  
  digitalWrite(led_green, LOW);
  tone(buzzer, 1200, 250); 
  delay(100);  
  for(int i=0;i<=2;i++) {
    digitalWrite(led_green, HIGH);
    delay(150);
    digitalWrite(led_green, LOW);
    delay(150);
  }
}

void ntpSyncDS1307() {
  sendNTPpacket(timeServer); // send an NTP packet to a time server
  // wait to see if a replay is available
  delay(100);
  if (Udp.parsePacket()) {
    // We've received a packet, read the data from it
    Udp.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
    // the timstamp starts at byte 40 of the received packet and is four bytes,
    // or two words, long. First, extract the two words:
    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
    // combine the four bytes (two words) into a long integer
    // this is NTP time (seconds since Jan 1 1900)
    unsigned long secsSince1900 = highWord << 16 | lowWord;
    // now convert NTP time into everyday time:
    // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
    const unsigned long seventyYears = 2208988800L;
    // substract seventy years:
    unsigned long epoch = secsSince1900 - seventyYears + timeZoneOffset + syncLatency;
    //Serial.println(epoch);
    setTime(epoch);
    RTC.set(epoch);
    ntpLastUpdate = now();
    // clear seconds displayed once ntp sync succeeded
    lcd.setCursor(9, 1);
    lcd.print(F("       "));
  }
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address) {
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011; // LI, Version, Mode
  packetBuffer[1] = 0; // Stratum, or type of clodk
  packetBuffer[2] = 6; // Polling Interval
  packetBuffer[3] = 0xEC; // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123);
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}

void digitalClockDisplay(){
  // digital clock display of the time
  dateDisplay();
  weekdayDisplay();
  timeDisplay();
  // display seconds since last NTP update
  lcd.setCursor(9, 1);
  lcd.print(now()-ntpLastUpdate);
}

void dateDisplay() {
  lcd.setCursor(0,0);
  lcd.print(year(t));
  lcd.print('-');
  if (month(t) < 10) {
    lcd.print('0');
  }
  lcd.print(month(t));
  lcd.print('-');
  if (day(t) < 10) {
    lcd.print('0');
  }
  lcd.print(day(t));
  // lcd.print(' ');
}

void weekdayDisplay() {
  lcd.setCursor(11,0);
  lcd.print(dayOfWeek[weekday()]);
}

void timeDisplay() {
  lcd.setCursor(0,1);
  if (hour(t) < 10) {
    lcd.print('0');
  }
  lcd.print(hour(t));
  lcd.print(':');
  if (minute(t) < 10) {
    lcd.print('0');
  }
  lcd.print(minute(t));
  lcd.print(':');
  if (second(t) < 10) {
    lcd.print('0');
  }
  lcd.print(second(t));
  // lcd.print(' ');
}

String setDataop() {
  String dtop;
  if (day(t) < 10) 
    dtop += "0";
    dtop += day(t);
  if (month(t) < 10) 
    dtop += "0";
    dtop += month(t);  
    dtop += year(t)-2000;
  if (hour(t) < 10) 
    dtop += "0";
    dtop += hour(t);
  if (minute(t) < 10) 
    dtop += "0";
    dtop += minute(t);
  return dtop;
  }

  
//---------------------------------------------------------------------------
// importo il file di configurazione ed aggiorno la eeprom 
//---------------------------------------------------------------------------

void loadConfiguration() {
  fileconfig = SD.open(configfilename,FILE_READ);
  if (fileconfig) {
  
    StaticJsonDocument<700> doc;
    DeserializationError error = deserializeJson(doc, fileconfig);
    if (error){
      Serial.println(F("Errore nel file di configurazione. Leggo i valori dalla eeprom"));
      loadEepromDefaults(); 
    }
    else {

      // ------------------------------------ ee_DB ----------------------------------------
      if (doc.containsKey("ee_db")) { 
        ee_db = doc["ee_db"];
        Serial.print(F("ee_db from config: "));
        Serial.println(ee_db);
      }
      else {
        ee_db = EEPROM.read(EEPROM_DB);
        Serial.print(F("ee_db from eeprom: "));
        Serial.println(ee_db);
      }
              
      if(ee_db != EEPROM.read(EEPROM_DB)) {
        EEPROM.write(EEPROM_DB,ee_db);
        Serial.print(F("ee_db - EEprom aggiornata: "));   
        Serial.println(EEPROM.read(EEPROM_DB));
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("EE_DB aggiornato"));
        lcd.setCursor(0,1);
        lcd.print(ee_db);
        delay(500);
      }
       
      // ------------------------------------ ee_SD ----------------------------------------
      if (doc.containsKey("ee_sd")) {
        ee_sd = doc["ee_sd"];
        Serial.print(F("ee_sd from config: "));
        Serial.println(ee_sd);
      }
      else {
        ee_sd = EEPROM.read(EEPROM_SD);
        Serial.print(F("ee_sd from eeprom: "));
        Serial.println(ee_sd);
      }
      if(ee_sd != EEPROM.read(EEPROM_SD))
      {
        EEPROM.write(EEPROM_SD,ee_sd);  
        Serial.print(F("ee_sd - EEprom aggiornata: "));   
        Serial.println(EEPROM.read(EEPROM_SD));
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("EE_SD aggiornato"));
        lcd.setCursor(0,1);
        lcd.print(ee_sd);
        delay(500);
      }
     
      // ------------------------------------ check utente ----------------------------------------
      if (doc.containsKey("ee_chk_ut")) 
      { 
        ee_chk_ut = doc["ee_chk_ut"];
        Serial.print(F("ee_chk_ut from config: "));
        Serial.println(ee_chk_ut);
      }
      else {
        ee_chk_ut = EEPROM.read(EEPROM_CHK_UT);
        Serial.print(F("ee_chk_ut from eeprom: "));
        Serial.println(ee_chk_ut);
      }
      if(ee_chk_ut != EEPROM.read(EEPROM_CHK_UT)) {
        EEPROM.write(EEPROM_CHK_UT,ee_chk_ut); 
        Serial.print(F("ee_chk_ut - EEprom aggiornata: "));   
        Serial.println(EEPROM.read(EEPROM_CHK_UT));
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("EE_chk_ut aggior"));
        lcd.setCursor(0,1);
        lcd.print(ee_chk_ut);
        delay(500);
      }
      
      // ------------------------------------ HEXDEC ----------------------------------------
      if (doc.containsKey("ee_hex_dec")) { 
        ee_hex_dec = doc["ee_hex_dec"];
        Serial.print(F("ee_hex_dec from config: "));
        Serial.println(ee_hex_dec);
      }
      else {
        ee_hex_dec = EEPROM.read(EEPROM_DEC_HEX);
        Serial.print(F("ee_hex_decde from eeprom: "));
        Serial.println(ee_hex_dec); 
      }
      if(ee_hex_dec != EEPROM.read(EEPROM_DEC_HEX)) {
        EEPROM.write(EEPROM_DEC_HEX,ee_hex_dec);  
        Serial.print(F("ee_hex_dec - EEprom aggiornata: "));   
        Serial.println(EEPROM.read(EEPROM_DEC_HEX));
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("EE_hexdec agg..."));
        lcd.setCursor(0,1);
        lcd.print(ee_hex_dec);
        delay(500);
      }
      
      // ------------------------------------ SEDE -----------------------------------------
      if (doc.containsKey("ee_sede")) {
        ee_sede = doc["ee_sede"];
        Serial.print(F("ee_sede from config: "));
        Serial.println(ee_sede);        
      }
      else {
        ee_sede = EEPROM.read(EEPROM_SEDE);
        Serial.print(F("ee_sede from eeprom: "));
        Serial.println(ee_sede);    
      }
      if(ee_sede != EEPROM.read(EEPROM_SEDE)) {
        EEPROM.write(EEPROM_SEDE,ee_sede);  
        Serial.print(F("ee_sede - EEprom aggiornata: "));   
        Serial.println(EEPROM.read(EEPROM_SEDE));
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("EE_sede aggiorn."));
        lcd.setCursor(0,1);
        lcd.print(ee_sede);
        delay(500);
      }

      // -------------------------------------- PING SERVER --------------------------------
      if (doc.containsKey("serverping")) {
        strlcpy(serverPingName, doc["serverping"], sizeof(serverPingName));
        Serial.print(F("serverPingName from config: "));
        for(int i = 0; i < strlen(serverPingName); i++){  Serial.print(serverPingName[i]);  }
      }
      else { 
        String retrievedString = readStringFromEEPROM(EEPROM_SERVERPINGNAME_ST);  // il primo byte riporta la lunghezza della stringa              
        for(int i = 0; i <retrievedString.length(); i++)  serverPingName[i] = retrievedString[i];       // 38 byte totali a disposizione = stringa max38 byte 
        Serial.print(F("serverPingName from eeprom: "));
        for(int i = 0; i < strlen(serverPingName); i++){  Serial.print(serverPingName[i]);  }
      }
      Serial.println();
      bool diverso = false;
      String retrievedString = readStringFromEEPROM(EEPROM_SERVERPINGNAME_ST);       
      for(int i=0; i<retrievedString.length(); i++) {  if (serverPingName[i]!=retrievedString[i])  diverso = true;  }    
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_SERVERPINGNAME_ST, serverPingName);
        retrievedString = readStringFromEEPROM(EEPROM_SERVERPINGNAME_ST); 
        Serial.print(F("serverPingName_st - EEprom aggiornata: "));   
        Serial.println(retrievedString);
        diverso = false;
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("ping svr aggiorn"));
        lcd.setCursor(0,1);
        lcd.print(retrievedString);
        delay(500);        
      }      
     
      // ----------------------------------------- SERVERCOMMAND ---------------------------------
      if (doc.containsKey("servercommand")) {
        strlcpy(serverCommandName, doc["servercommand"], sizeof(serverCommandName));
        Serial.print(F("serverCommandName from config: "));
        for(int i = 0; i < strlen(serverCommandName); i++){  Serial.print(serverCommandName[i]);  }
      }
      else { 
        String retrievedString = readStringFromEEPROM(EEPROM_SERVERCOMMANDNAME_ST);  // il primo byte riporta la lunghezza della stringa              
        for(int i = 0; i <retrievedString.length(); i++)  serverCommandName[i] = retrievedString[i];       // 38 byte totali a disposizione = stringa max38 byte 
        Serial.print(F("serverCommandName from eeprom: "));
        for(int i = 0; i < strlen(serverCommandName); i++){  Serial.print(serverCommandName[i]);  }
      }
      Serial.println();
      diverso = false;
      retrievedString = readStringFromEEPROM(EEPROM_SERVERCOMMANDNAME_ST);       
      for(int i=0; i<retrievedString.length(); i++) {  if (serverCommandName[i]!=retrievedString[i])  diverso = true;  }    
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_SERVERCOMMANDNAME_ST, serverCommandName);
        retrievedString = readStringFromEEPROM(EEPROM_SERVERCOMMANDNAME_ST); 
        Serial.print(F("serverCommandName_st - EEprom aggiornata: "));   
        Serial.println(retrievedString);
        diverso = false;
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("command svr agg."));
        lcd.setCursor(0,1);
        lcd.print(retrievedString);
        delay(500);     
      } 
      // ------------------------------------------ inizio ntp sync interval -----------------------
      if (doc.containsKey("ntp_sync_interval")) 
      {
        ntpSyncInterval = doc["ntp_sync_interval"];
        Serial.print(F("ntpSyncInterval from config: "));
        Serial.println(ntpSyncInterval);
      }
      else{
        String ntpSyncInterval_ee_str = readStringFromEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST); 
        ntpSyncInterval = ntpSyncInterval_ee_str.toInt();
        Serial.print(F("ntpSyncInterval from eeprom: "));
        Serial.println(ntpSyncInterval);
      }      
      diverso = false;
      if(ntpSyncInterval != readStringFromEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST).toInt())
        diverso = true;
                 
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST, String(ntpSyncInterval));
        String ntpSyncInterval_ee_str = readStringFromEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST); 
        Serial.print(F("ntpSyncInterval_ee_str - EEprom aggiornata : "));
        Serial.println(ntpSyncInterval_ee_str); 
        diverso = false; 
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("NTP tm_sync agg."));
        lcd.setCursor(0,1);
        lcd.print(ntpSyncInterval_ee_str);
        delay(500);           
      } 
      
      // ----------------------------------------------- urlpost -------------------------------
      if (doc.containsKey("urlpost")) {
        char url[39]="";
        strlcpy(url, doc["urlpost"], sizeof(url));
        urlPost = url;
        Serial.print(F("urlPost from config: "));
        Serial.print(urlPost);  
      }
      else { 
        String retrievedString = readStringFromEEPROM(EEPROM_URLPOST_ST);  // il primo byte riporta la lunghezza della stringa              
        urlPost = retrievedString;                                                   // 38 byte totali a disposizione = stringa max38 byte 
        Serial.print(F("urlPost from eeprom: "));
        Serial.print(urlPost); 
      }
      Serial.println();
      diverso = false;
      retrievedString = readStringFromEEPROM(EEPROM_URLPOST_ST);       
      if (urlPost!=retrievedString)  diverso = true;      
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_URLPOST_ST, urlPost);
        retrievedString = readStringFromEEPROM(EEPROM_URLPOST_ST); 
        Serial.print(F("urlPost_st - EEprom aggiornata: "));   
        Serial.println(retrievedString);
        diverso = false;
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("urlPost aggiorn."));
        lcd.setCursor(0,1);
        lcd.print(retrievedString);
        delay(500); 
      } 
      // ------------------------------------------ inizio serverCommandPort --------------------
      if (doc.containsKey("servercommandport")) 
      {
        serverCommandPort = doc["servercommandport"];
        Serial.print(F("serverCommandPort from config: "));
        Serial.println(serverCommandPort);
      }
      else{
        String serverCommandPort_ee_str = readStringFromEEPROM(EEPROM_SERVERCOMMANDPORT_ST); 
        serverCommandPort = serverCommandPort_ee_str.toInt();
        Serial.print(F("serverCommandPort from eeprom: "));
        Serial.println(serverCommandPort);
      }      
      diverso = false;
      if(serverCommandPort != readStringFromEEPROM(EEPROM_SERVERCOMMANDPORT_ST).toInt())
        diverso = true;
                 
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_SERVERCOMMANDPORT_ST, String(serverCommandPort));
        String serverCommandPort_ee_str = readStringFromEEPROM(EEPROM_SERVERCOMMANDPORT_ST); 
        Serial.print(F("serverCommandPort_ee_str - EEprom aggiornata : "));
        Serial.println(serverCommandPort_ee_str); 
        diverso = false;  
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("svrCommand agg.."));
        lcd.setCursor(0,1);
        lcd.print(serverCommandPort_ee_str);
        delay(500);      
      }       
            
      // -------------------------------------- inizio ip -------------------------------------
      if (doc.containsKey("ip")) 
      {
        String ip_str = doc["ip"];
        ip.fromString(ip_str);
        Serial.print(F("ip from config: "));
        Serial.println(ip);         
      }
      else{
        String ip_str = readStringFromEEPROM(EEPROM_IP_ST);
        ip.fromString(ip_str);
        Serial.print(F("ip from eeprom: "));
        Serial.println(ip);
      }      
      diverso = false;
      IPAddress iptmp;
      String ip_str2 = readStringFromEEPROM(EEPROM_IP_ST);
      iptmp.fromString(ip_str2);      
      if(ip != iptmp) diverso = true;     
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_IP_ST, String(ip[0]) + '.' + String(ip[1]) + '.' + String(ip[2]) + '.' + String(ip[3])); // il contrario);
        String ip_str = readStringFromEEPROM(EEPROM_IP_ST);
        ip.fromString(ip_str); 
        Serial.print(F("IP - EEprom aggiornata : "));
        Serial.println(ip); 
        diverso = false;  
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("indir. IP agg.OK"));
        lcd.setCursor(0,1);
        lcd.print(ip);
        delay(500);             
      }       
      
      // ------------------------------------------- inizio subnet --------------------------
      if (doc.containsKey("subnet")) 
      {
        String sub_str = doc["subnet"];
        subnet.fromString(sub_str);
        Serial.print(F("subnet from config: "));
        Serial.println(subnet);           
      }
      else{
        String sub_str = readStringFromEEPROM(EEPROM_SUB_ST);
        subnet.fromString(sub_str);
        Serial.print(F("subnet from eeprom: "));
        Serial.println(subnet);
      }      
      diverso = false;
      IPAddress subtmp;
      String sub_str2 = readStringFromEEPROM(EEPROM_SUB_ST);
      subtmp.fromString(sub_str2);      
      if(subnet != subtmp) diverso = true;     
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_SUB_ST, String(subnet[0]) + '.' + String(subnet[1]) + '.' + String(subnet[2]) + '.' + String(subnet[3])); // il contrario);
        String sub_str = readStringFromEEPROM(EEPROM_SUB_ST);
        subnet.fromString(sub_str); 
        Serial.print(F("subnet - EEprom aggiornata : "));
        Serial.println(subnet); 
        diverso = false; 
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("subnet aggiorn.."));
        lcd.setCursor(0,1);
        lcd.print(subnet);
        delay(500);      
      }       
      
      // ----------------------------------- gateway ---------------------------------------
      if (doc.containsKey("gateway")) 
      {
        String gateway_str = doc["gateway"];
        gateway.fromString(gateway_str);
        Serial.print(F("gateway from config: "));
        Serial.println(gateway);           
      }
      else{
        String gateway_str = readStringFromEEPROM(EEPROM_GAT_ST);
        gateway.fromString(gateway_str);
        Serial.print(F("gateway from eeprom: "));
        Serial.println(gateway);
      }      
      diverso = false;
      IPAddress gatewaytmp;
      String gateway_str2 = readStringFromEEPROM(EEPROM_GAT_ST);
      gatewaytmp.fromString(gateway_str2);      
      if(gateway != gatewaytmp) diverso = true;     
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_GAT_ST, String(gateway[0]) + '.' + String(gateway[1]) + '.' + String(gateway[2]) + '.' + String(gateway[3])); // il contrario);
        String gateway_str = readStringFromEEPROM(EEPROM_GAT_ST);
        gateway.fromString(gateway_str); 
        Serial.print(F("Gateway - EEprom aggiornata : "));
        Serial.println(gateway); 
        diverso = false;  
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("gateway aggiorn."));
        lcd.setCursor(0,1);
        lcd.print(gateway);
        delay(500);      
      }       
      
      // --------------------------------- DNS -----------------------------------------------
      if (doc.containsKey("dns")) 
      {
        String dns_str = doc["dns"];
        dnServer.fromString(dns_str);
        Serial.print(F("DNS from config: "));
        Serial.println(dnServer);           
      }
      else{
        String dns_str = readStringFromEEPROM(EEPROM_DNS_ST);
        dnServer.fromString(dns_str);
        Serial.print(F("DNS from eeprom: "));
        Serial.println(dnServer);
      }      
      diverso = false;
      IPAddress dnServer_tmp;
      String dnServer_str2 = readStringFromEEPROM(EEPROM_DNS_ST);
      dnServer_tmp.fromString(dnServer_str2);      
      if(dnServer != dnServer_tmp) diverso = true;     
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_DNS_ST, String(dnServer[0]) + '.' + String(dnServer[1]) + '.' + String(dnServer[2]) + '.' + String(dnServer[3])); // il contrario);
        String dnServer_str = readStringFromEEPROM(EEPROM_DNS_ST);
        dnServer.fromString(dnServer_str); 
        Serial.print(F("DNS - EEprom aggiornata : "));
        Serial.println(dnServer); 
        diverso = false;   
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("DNS svr aggiorn."));
        lcd.setCursor(0,1);
        lcd.print(dnServer);
        delay(500);    
      }      
      
      // ---------------------------------- inizio ntpsvr --------------------------------
      if (doc.containsKey("ntpsvr")) 
      {
        String ntpsvr_str = doc["ntpsvr"];
        timeServer.fromString(ntpsvr_str);
        Serial.print(F("NTP from config: "));
        Serial.println(timeServer);           
      }
      else{
        String ntpsvr_str = readStringFromEEPROM(EEPROM_NTP_ST);
        timeServer.fromString(ntpsvr_str);
        Serial.print(F("NTP from eeprom: "));
        Serial.println(timeServer);
      }      
      diverso = false;
      IPAddress timeServer_tmp;
      String timeServer_str2 = readStringFromEEPROM(EEPROM_NTP_ST);
      timeServer_tmp.fromString(timeServer_str2);      
      if(timeServer != timeServer_tmp) diverso = true;     
      if (diverso==true)
      {       
        writeStringToEEPROM(EEPROM_NTP_ST, String(timeServer[0]) + '.' + String(timeServer[1]) + '.' + String(timeServer[2]) + '.' + String(timeServer[3])); // il contrario);
        String timeServer_str = readStringFromEEPROM(EEPROM_NTP_ST);
        timeServer.fromString(timeServer_str); 
        Serial.print(F("NTPsvr - EEprom aggiornata : "));
        Serial.println(timeServer); 
        diverso = false;
        lcd.clear();
        lcd.setCursor(0,0);
        lcd.println(F("TimeSvr aggiorn."));
        lcd.setCursor(0,1);
        lcd.print(timeServer);
        delay(500);       
      }       
      //  ------------------ fine Timeserver -----------------------
                  
    }      
      Serial.println(F("Valori aggiornati dal file config."));
      fileconfig.close();       
  }    
  else {
    Serial.println(F("Errore SD o SD non inserita. Leggo i valori dalla eeprom"));
    loadEepromDefaults();
  }  
}

//---------------------------------------------------------------
int writeStringToEEPROM(int addrOffset, const String &strToWrite)
{
  byte len = strToWrite.length();
  EEPROM.update(addrOffset, len);
  for (int i = 0; i < len; i++)
    EEPROM.update(addrOffset + 1 + i, strToWrite[i]);
  return addrOffset + 1 + len;
}

//---------------------------------------------------------------
String readStringFromEEPROM(int addrOffset)
{
  int newStrLen = EEPROM.read(addrOffset);
  char data[newStrLen + 1];
  for (int i = 0; i < newStrLen; i++)
    data[i] = EEPROM.read(addrOffset + 1 + i);
    
  data[newStrLen] = '\0'; // the character may appear in a weird way, you should read: 'only one backslash and 0'
  return String(data);
}

//-----------------------------------------
// Leggo i parametri dalla eeprom
//-----------------------------------------
void loadEepromDefaults(){
  ee_db = EEPROM.read(EEPROM_DB);
  ee_sd = EEPROM.read(EEPROM_SD);
  ee_chk_ut = EEPROM.read(EEPROM_CHK_UT);
  ee_hex_dec = EEPROM.read(EEPROM_DEC_HEX);
  ee_sede = EEPROM.read(EEPROM_SEDE);    
  sede3ch="00";
  sede3ch.concat(ee_sede);  
  Serial.print(F("db:"));
  Serial.print(ee_db);
  Serial.print(F("  sd:"));
  Serial.print(ee_sd);
  Serial.print(F("  chk_ut:"));
  Serial.print(ee_chk_ut);
  Serial.print(F("  ee_hex_dec:"));
  Serial.print(ee_hex_dec);
  Serial.print(F("  sede:"));
  Serial.println(sede3ch);
  
  String serverPingName_st = readStringFromEEPROM(EEPROM_SERVERPINGNAME_ST); 
  for(int i = 0; i < serverPingName_st.length(); i++) serverPingName[i] = serverPingName_st[i];
  Serial.print(F("PingSvr load: "));
  for(int i = 0; i < strlen(serverPingName); i++)  Serial.print(serverPingName[i]);
  Serial.println();  

  String serverCommandName_st = readStringFromEEPROM(EEPROM_SERVERCOMMANDNAME_ST); 
  for(int i = 0; i < serverCommandName_st.length(); i++) serverCommandName[i] = serverCommandName_st[i];
  Serial.print(F("serverCommandName load: "));
  for(int i = 0; i < strlen(serverCommandName); i++)  Serial.print(serverCommandName[i]);
  Serial.println();
  
  String ntpSyncInterval_ee_str = readStringFromEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST); 
  ntpSyncInterval = ntpSyncInterval_ee_str.toInt();
  Serial.print(F("ntpSyncInterval load: "));
  Serial.println(ntpSyncInterval);
  
  urlPost = readStringFromEEPROM(EEPROM_URLPOST_ST); 
  Serial.print(F("urlPost load: "));
  Serial.println(urlPost);

  String serverCommandPort_ee_str = readStringFromEEPROM(EEPROM_SERVERCOMMANDPORT_ST); 
  serverCommandPort = serverCommandPort_ee_str.toInt();
  Serial.print(F("serverCommandPort load: "));
  Serial.println(serverCommandPort);

  String ip_str = readStringFromEEPROM(EEPROM_IP_ST); 
  ip.fromString(ip_str);
  Serial.print(F("ip load: "));
  Serial.println(ip);

  String sub_str = readStringFromEEPROM(EEPROM_SUB_ST); 
  subnet.fromString(sub_str);
  Serial.print(F("subnet load: "));
  Serial.println(subnet);

  String gat_str = readStringFromEEPROM(EEPROM_GAT_ST); 
  gateway.fromString(gat_str);
  Serial.print(F("gateway load: "));
  Serial.println(gateway);

  String dns_str = readStringFromEEPROM(EEPROM_DNS_ST); 
  dnServer.fromString(dns_str);
  Serial.print(F("DNS load: "));
  Serial.println(dnServer);
  
  String ntp_str = readStringFromEEPROM(EEPROM_NTP_ST); 
  timeServer.fromString(ntp_str);
  Serial.print(F("NTP load: "));
  Serial.println(timeServer);
  
}

void inizializza_defaults(){
  EEPROM.update(EEPROM_DB,1);         // indirizzo 0: Abilito scrittura su DB
  EEPROM.update(EEPROM_SD,1);         // indirizzo 1: Abilito scrittura su SD card
  EEPROM.update(EEPROM_CHK_UT,0);     // check dell'esistenza dell'utente su db 
  EEPROM.update(EEPROM_DEC_HEX,0);    // Switch dei tag tra dec e hex
  EEPROM.update(EEPROM_SEDE,5);       // codice Sede 
  
  writeStringToEEPROM(EEPROM_SERVERPINGNAME_ST, "www.google.it");       // svr ping di default
  String default_string = readStringFromEEPROM(EEPROM_SERVERPINGNAME_ST); 
  Serial.print(F("PingSvr init: "));
  Serial.println(default_string);
  
  writeStringToEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST, "3600");  
  default_string = readStringFromEEPROM(EEPROM_NTP_SYNC_INTERVAL_ST); 
  Serial.print(F("ntpSyncInterval init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_SERVERCOMMANDNAME_ST, "www.thepiper.eu");    
  default_string = readStringFromEEPROM(EEPROM_SERVERCOMMANDNAME_ST); 
  Serial.print(F("serverCommandName init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_URLPOST_ST, "/meteo/timbratura.php");    
  default_string = readStringFromEEPROM(EEPROM_URLPOST_ST); 
  Serial.print(F("urlPost init: "));
  Serial.println(default_string);
  
  writeStringToEEPROM(EEPROM_SERVERCOMMANDPORT_ST, "80");    
  default_string = readStringFromEEPROM(EEPROM_SERVERCOMMANDPORT_ST); 
  Serial.print(F("serverCommandPort init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_IP_ST, "10.10.10.200");    
  default_string = readStringFromEEPROM(EEPROM_IP_ST); 
  Serial.print(F("Ip init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_SUB_ST, "255.255.255.0");    
  default_string = readStringFromEEPROM(EEPROM_SUB_ST); 
  Serial.print(F("Subnet init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_GAT_ST, "10.10.10.254");    
  default_string = readStringFromEEPROM(EEPROM_GAT_ST); 
  Serial.print(F("Gateway init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_DNS_ST, "8.8.8.8");    
  default_string = readStringFromEEPROM(EEPROM_DNS_ST); 
  Serial.print(F("DNS init: "));
  Serial.println(default_string);

  writeStringToEEPROM(EEPROM_NTP_ST, "162.159.200.1");    
  default_string = readStringFromEEPROM(EEPROM_DNS_ST); 
  Serial.print(F("NTP init: "));
  Serial.println(default_string);

  
  
}

